﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Runtime.InteropServices;

using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using VRage;
using VRage.Game.Components;
using System.Xml.Serialization;
using VRage.Serialization;
using Draygo.API;
using VRageMath;
using BlendTypeEnum = VRageRender.MyBillboard.BlendTypeEnum;

// This handles server administrative functions
// Some of this code was take from Midspace's Admin helper
namespace Phoenix.AutoOreCollect
{
    public enum OnOffTriState
    {
        On,
        Off,
        Unset
    }

    public static class Globals
    {
        public const string ModName = "AutoOre";
    }

    public enum OreCollectionMethod
    {
        Collect,
        NoCollect,
        Destroy,
    }

    /// <summary>
    /// This is the mod configuration data stored on disk
    /// </summary>
    public class PlayerOptions
    {
        public bool HandPickup = true;
        public bool HUDEnabled = true;
        public SerializableDictionary<string, OreCollectionMethod> OreCollectionMethod = new SerializableDictionary<string, OreCollectionMethod>();

        public PlayerOptions()
        {
            HandPickup = AutoOreAdmin.Configuration.Enabled;
        }
    }

    [XmlInclude(typeof(PlayerOptions))]
    [XmlInclude(typeof(OreCollectionMethod))]
    public class AutoOreConfig
    {
        public bool Debug = false;
        public bool Enabled = true;
        public float RangeMultiplier = 2.0f;
        public SerializableDictionary<ulong, PlayerOptions> PlayerData = new SerializableDictionary<ulong, PlayerOptions>();
        public SerializableDictionary<string, OreCollectionMethod> OreCollection = new SerializableDictionary<string, OreCollectionMethod>();

        public bool IsEnabled(ulong playerSteamId)
        {
            return PlayerData.Dictionary.GetValueOrDefault(playerSteamId, new PlayerOptions()).HandPickup;
        }

        public bool IsEnabled()
        {
            return PlayerData.Dictionary.GetValueOrDefault(MyAPIGateway.Session?.Player?.SteamUserId ?? 0, new PlayerOptions()).HandPickup;
        }

        public bool IsHUDEnabled(ulong playerSteamId)
        {
            return PlayerData.Dictionary.GetValueOrDefault(playerSteamId, new PlayerOptions()).HUDEnabled;
        }

        public bool IsHUDEnabled()
        {
            return PlayerData.Dictionary.GetValueOrDefault(MyAPIGateway.Session?.Player?.SteamUserId ?? 0, new PlayerOptions()).HUDEnabled;
        }

        public void SetEnabled(ulong playerSteamId, bool enabled)
        {
            var data = PlayerData.Dictionary.GetValueOrDefault(playerSteamId, new PlayerOptions());
            data.HandPickup = enabled;
            PlayerData[playerSteamId] = data;
        }

        public void SetHUDEnabled(ulong playerSteamId, bool enabled)
        {
            var data = PlayerData.Dictionary.GetValueOrDefault(playerSteamId, new PlayerOptions());
            data.HUDEnabled = enabled;
            PlayerData[playerSteamId] = data;

            if (playerSteamId == MyAPIGateway.Session?.Player?.SteamUserId)
                AutoOreAdmin.Instance.ShowHUD = enabled;
        }

        public OreCollectionMethod GetCollectionMethod(ulong playerSteamId, string orename)
        {
            return PlayerData.Dictionary.GetValueOrDefault(                             // Get player value
                    playerSteamId, new PlayerOptions()).                                // Or fake it with new data
                        OreCollectionMethod.Dictionary.GetValueOrDefault(               // Get Collection method from player
                            orename, OreCollection.Dictionary.GetValueOrDefault(        // Or get server value if not exist
                                orename, IsEnabled(playerSteamId) ? OreCollectionMethod.Collect : OreCollectionMethod.NoCollect));  // Default to global setting
        }

        public void SetCollectionMethod(ulong playerSteamId, string orename, OreCollectionMethod method)
        {
            var playerdata = PlayerData.Dictionary.GetValueOrDefault(playerSteamId, new PlayerOptions());
            playerdata.OreCollectionMethod.Dictionary[orename] = method;
            PlayerData.Dictionary[playerSteamId] = playerdata;
        }
    }

    public enum CollectionState
    {
        NoDrill,
        Off,
        Ready,
        Collecting,
    }

    [MySessionComponentDescriptor(MyUpdateOrder.BeforeSimulation)]
    class AutoOreAdmin : MySessionComponentBase
    {
        #region Static data
        static bool _isInitialized = false;
        public static AutoOreAdmin Instance { get; private set; }
        static AutoOreConfig _config = null;
        public static AutoOreConfig Configuration
        {
            get
            {
                if (_config == null)
                    _config = new AutoOreConfig();
                return _config;
            }
            private set
            {
                _config = value;
            }
        }
        #endregion

        #region HUD API
        HudAPIv2.HUDMessage m_status2;
        StringBuilder m_message = new StringBuilder();
        HudAPIv2 m_HudAPI;
        public HudAPIv2 HUDApi { get { return m_HudAPI; } }
        public bool ShowHUD { get; set; }

        #region Menu
        private HudAPIv2.MenuRootCategory ClientSettingsCategory;
        private HudAPIv2.MenuRootCategory AdminSettingsCategory;
        private HudAPIv2.MenuScreenInput NotificationsPosition;
        private HudAPIv2.MenuItem SettingsEnable;
        private HudAPIv2.MenuItem SettingsServerEnable;
        private HudAPIv2.MenuItem SettingsShowHUD;

        private HudAPIv2.MenuItem Separator1;
        private HudAPIv2.MenuItem SaveSettings;
        private HudAPIv2.MenuItem Separator2;
        private HudAPIv2.MenuItem ResetSettings;
        #endregion

        #endregion HUD API

        public CollectionState CollectionState { get; set; }
        private const string ConfigFileName = "Config_{0}.cfg";

        static AutoOreAdmin()
        {
            _config = new AutoOreConfig();
        }

        public AutoOreAdmin()
        {
            Instance = this;
        }

        public override void Draw()
        {
            base.Draw();
            string message = string.Empty;

            switch(CollectionState)
            {
                case CollectionState.Off:
                    message = "<Color=Red>" + CollectionState.ToString();
                    break;
                case CollectionState.NoDrill:
                    message = "<Color=Yellow>No Drill";
                    break;
                case CollectionState.Ready:
                    message = "<Color=Green>" + CollectionState.ToString();
                    break;
                case CollectionState.Collecting:
                    message = "<Color=Cyan>" + CollectionState.ToString();
                    break;
            }

            m_message.Clear();
            m_message.Append("Ore Collection: ");
            m_message.Append(message);

            if (m_status2 != null && ShowHUD != m_status2.Visible)
                m_status2.Visible = ShowHUD;
        }

        private void UpdateLabels()
        {
            SettingsEnable.Text = string.Format("Enabled: {0}", Configuration.IsEnabled() ? "On" : "Off");
            SettingsServerEnable.Text = string.Format("Enabled: {0}", Configuration.Enabled ? "On" : "Off");
            SettingsShowHUD.Text = string.Format("Show HUD: {0}", Configuration.IsHUDEnabled() ? "On" : "Off");
        }

        private void Init()
        {
            _isInitialized = true;
            LoadConfig();
            Logger.Instance.LogDebug("AutoOreAdmin.Init()");
            MyAPIGateway.Utilities.MessageEntered += Utilities_MessageEntered;
            MyAPIGateway.Multiplayer.RegisterMessageHandler(MessageUtils.MessageId, MessageUtils.HandleMessage);
            m_HudAPI = new HudAPIv2(() =>
            {
                if (m_HudAPI.Heartbeat)
                {
                    Vector2D DefaultPosition = new Vector2D(0.72, -0.50);

                    m_status2 = new HudAPIv2.HUDMessage(m_message, DefaultPosition, HideHud: true, Shadowing: true, ShadowColor: Color.Black, Blend: BlendTypeEnum.PostPP);
                    ShowHUD = m_status2.Visible = Configuration.IsHUDEnabled();

                    // Setup menu
                    ClientSettingsCategory = new HudAPIv2.MenuRootCategory("Automatic Ore Collection", HudAPIv2.MenuRootCategory.MenuFlag.PlayerMenu, "Auto-Ore Settings");
                    AdminSettingsCategory = new HudAPIv2.MenuRootCategory("Automatic Ore Collection (Admin)", HudAPIv2.MenuRootCategory.MenuFlag.AdminMenu, "Auto-Ore Settings (Admin)");

                    SettingsEnable = new HudAPIv2.MenuItem("Enabled", ClientSettingsCategory, () =>
                    {
                        MessageUtils.SendMessageToServer(new MessagePlayerToggle() { Enable = !Configuration.IsEnabled() });
                        UpdateLabels();
                    });

                    SettingsServerEnable = new HudAPIv2.MenuItem("Enabled", AdminSettingsCategory, () =>
                    {
                        Configuration.Enabled = !Configuration.Enabled;
                        UpdateLabels();
                    });

                    SettingsShowHUD = new HudAPIv2.MenuItem("Show HUD", ClientSettingsCategory, () =>
                    {
                        var message = new MessageHUD() { Enable = !Configuration.IsHUDEnabled() };
                        MessageUtils.SendMessageToServer(message);
                        UpdateLabels();
                    });

                    Separator1 = new HudAPIv2.MenuItem(null, ClientSettingsCategory, Interactable: false);

                    SaveSettings = new HudAPIv2.MenuItem("Save Settings", ClientSettingsCategory, () =>
                    {
                        SaveConfig(true);
                    });

                    Separator2 = new HudAPIv2.MenuItem(null, ClientSettingsCategory, Interactable: false);

                    //ResetSettings = new HudAPIv2.MenuItem("Reset Settings", ClientSettingsCategory, () =>
                    //{
                    //});

                    UpdateLabels();
                }
            });
        }

        public override void LoadData()
        {
            Logger.Instance.Init("AutoOreCollect");
            Logger.Instance.LogDebug("AutoOreAdmin.LoadData()");
        }

        public override void UpdateBeforeSimulation()
        {
            if (!_isInitialized && MyAPIGateway.Session != null)
                Init();
        }

        protected override void UnloadData()
        {
            try
            {
                MyAPIGateway.Utilities.MessageEntered -= Utilities_MessageEntered;
                MyAPIGateway.Multiplayer.UnregisterMessageHandler(MessageUtils.MessageId, MessageUtils.HandleMessage);
                Logger.Instance.Close();
            }
            catch (Exception ex) { Logger.Instance.LogException(ex); }

        }

        public override void SaveData()
        {
            SaveConfig();
        }

        static void Utilities_MessageEntered(string messageText, ref bool sendToOthers)
        {
            if (ProcessClientMessage(messageText))
                sendToOthers = false;
        }

        public static bool ProcessClientMessage(string messageText)
        {
            if (!_isInitialized || string.IsNullOrEmpty(messageText))
                return false;

            string invalidOptionErrorText = "Invalid argument supplied. type /" + Globals.ModName.ToLowerInvariant() + " help to show valid commands and options";

            Logger.Instance.LogDebug("Processing message");

            var commands = messageText.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            if (commands.Length == 0)
                return false;

            try
            {
                var match = Regex.Match(messageText, @"(/" + Globals.ModName.ToLowerInvariant() + @")\s+(?<Key>[^\s]+)((\s+(?<Value>.+))|)", RegexOptions.IgnoreCase);
                if (match.Success)
                {
                    var key = match.Groups["Key"].Value;
                    var value = match.Groups["Value"].Value.Trim();

                    // Allowed client options
                    switch (key.ToLowerInvariant())
                    {
                        case "ore":
                        case "showhud":
                        case "hud":
                        case "help":
                            break;
                        default:
                            if (!MyAPIGateway.Session.Player.IsAdmin)
                            {
                                MyAPIGateway.Utilities.ShowMessage(Globals.ModName, "You must be a server admin to use this.");
                                return true;
                            }
                            break;
                    }

                    bool bval = false;
                    OnOffTriState onoffval = OnOffTriState.Unset;
                    float fval;

                    switch (key.ToLowerInvariant())
                    {
                        case "collect":
                            bval = true;
                            onoffval = OnOffTriState.Unset;
                            if (bool.TryParse(value, out bval) || OnOffTriState.TryParse(value, true, out onoffval))
                            {
                                if (onoffval != OnOffTriState.Unset)
                                    bval = onoffval == OnOffTriState.On ? true : false;

                                var message = new MessageEnable() { Enable = bval };
                                MessageUtils.SendMessageToServer(message);
                            }
                            else
                                MyAPIGateway.Utilities.ShowMessage(Globals.ModName, invalidOptionErrorText);
                            break;
                        case "showhud":
                        case "hud":
                            bval = true;
                            onoffval = OnOffTriState.Unset;
                            if (bool.TryParse(value, out bval) || OnOffTriState.TryParse(value, true, out onoffval))
                            {
                                if (onoffval != OnOffTriState.Unset)
                                    bval = onoffval == OnOffTriState.On ? true : false;

                                var message = new MessageHUD() { Enable = bval };
                                MessageUtils.SendMessageToServer(message);
                            }
                            else
                                MyAPIGateway.Utilities.ShowMessage(Globals.ModName, invalidOptionErrorText);
                            break;
                        case "rangemultiplier":
                        case "rangemult":
                        case "range":
                            fval = 0;
                            if (float.TryParse(value, out fval))
                            {
                                if (fval < 1.0 || fval > 100)
                                {
                                    MyAPIGateway.Utilities.ShowMessage(Globals.ModName, "Value out of range. Must be between 1-100.");
                                    return true;
                                }

                                var message = new MessageSetRange() { RangeMultiplier = fval };
                                MessageUtils.SendMessageToServer(message);
                            }
                            else
                            {
                                if (string.Compare(value, "reset", true) == 0)
                                    MessageUtils.SendMessageToServer(new MessageSetRange() { Reset = true });
                                else
                                    MyAPIGateway.Utilities.ShowMessage(Globals.ModName, invalidOptionErrorText);
                            }
                            break;
                        case "serverore":
                        case "ore":
                            // This sets the upgrade values (overrides cubeblocks.sbc)
                            if (value.ToLowerInvariant().Contains("help"))
                            {
                                MyAPIGateway.Utilities.ShowMissionScreen(Globals.ModName + " Admin Help",
                                    "/" + Globals.ModName.ToLowerInvariant() + " " + key.ToLowerInvariant() + " <OreName> <Method>|reset", "",
                                    "<Method> collection method for specific <OreName>:\r\n" +
                                    "Collect - Always collect this ore while drilling\r\n" +
                                    "Never - Never collect this ore while drilling (will be floating)\r\n" +
                                    "Destroy - Destroy this ore while drilling\r\n" +
                                    "\r\n" +
                                    "Example:\r\n" +
                                    "/autoore ore stone destroy\r\n" +
                                    "\r\n" +
                                    //"LargeRange, SmallRange, and InhibitorRange are in meters, all others are percentage as a decimal (0.5)\r\n" +
                                    "Use 'reset' as the Method to reset the value to the default." +
                                    ""
                                    );
                                break;
                            }
                            if (!HandleValueChange(value, string.Compare(key, "serverore", true) == 0))
                                MyAPIGateway.Utilities.ShowMessage(Globals.ModName, invalidOptionErrorText);
                            break;
                        case "debug":
                            bool debug = true;
                            OnOffTriState dbg = OnOffTriState.Unset;
                            if (bool.TryParse(value, out debug) || OnOffTriState.TryParse(value, true, out dbg))
                            {
                                if (dbg != OnOffTriState.Unset)
                                    debug = dbg == OnOffTriState.On ? true : false;

                                var message = new MessageDebug() { DebugMode = debug };
                                message.InvokeProcessing();
                                MessageUtils.SendMessageToServer(message);
                            }
                            else
                                MyAPIGateway.Utilities.ShowMessage(Globals.ModName, invalidOptionErrorText);
                            break;
                        case "save":
                            SaveConfig(true);
                            break;
                        case "help":
                            MyAPIGateway.Utilities.ShowMissionScreen(Globals.ModName + " Admin Help",
                                "/" + Globals.ModName.ToLowerInvariant() + " <command> [value]", "",
                                "Current commands:\r\n" +
                                "collect <on|off|true|false> - Enable/disable auto ore collection by default\r\n" +
                                "rangemultiplier <float> - Set collection range multiplier (multiplied by WorldVolume.Radius). Between 1-100.\r\n" +
                                "ore <orename> <collect|nocollect|destroy> - Set player option for specific ore collection\r\n" +
                                "serverore <orename> <collect|nocollect|destroy> - Set server option for specific ore collection\r\n" +
                                "showhud <on|off|true|false> - Show HUD text" + 
                                "debug <on|off|true|false> - Enable/disable debug logging\r\n" +
                                "save - Save settings\r\n" +
                                "\r\n" +
                                "use '/" + Globals.ModName.ToLowerInvariant() + " ore help' for details\r\n" +
                                "Some commands are server-wide and can only be run by server administrators.\r\n" +
                                "Items with [ ] are optional, < > are required.\r\n" +
                                "<on|off> means pick 'on' or 'off'.\r\n" +
                                "\r\n" +
                                "If problems occur, do '/" + Globals.ModName.ToLowerInvariant() + " save' and reload the map.\r\n" +
                                ""
                                );
                            break;
                        default:
                            MyAPIGateway.Utilities.ShowMessage(Globals.ModName, "Invalid command, Method /" + Globals.ModName.ToLowerInvariant() + " help to show valid commands");
                            break;
                    }
                    Instance.UpdateLabels();
                    return true;
                }
            }
            catch(Exception ex)
            {
                Logger.Instance.LogException(ex);
            }

            return false;
        }

        static bool HandleValueChange(string text, bool serveroption)
        {
            try
            {
                var match = Regex.Match(text, @"(?<OreName>[^\s]+)(\s+(?<OreCollectionMethod>collect|nocollect|destroy|reset))", RegexOptions.IgnoreCase);
                if (match.Success)
                {
                    var stype = match.Groups["OreName"].Value.Trim().ToLowerInvariant();
                    var modifier = match.Groups["OreCollectionMethod"].Value.Trim();
                    OreCollectionMethod Method = OreCollectionMethod.Collect;

                    if (!Enum.TryParse<OreCollectionMethod>(modifier, true, out Method))
                    {
                        if (modifier.ToLowerInvariant() == "reset")
                        {
                            MessageUtils.SendMessageToServer(new MessageValueChange() { OreName = stype, ServerOption = serveroption, Reset = true });
                            return true;
                        }
                        MyAPIGateway.Utilities.ShowMessage(Globals.ModName, "Invalid method: " + stype);
                        return false;
                    }

                    Logger.Instance.LogDebug(string.Format("OreName: {0}, OreCollectionMethod: {1}", stype, Method));
                    MessageUtils.SendMessageToServer(new MessageValueChange() { OreName = stype, Method = Method, ServerOption = serveroption });
                    return true;
                }
            }
            catch (Exception ex)
            {
                Logger.Instance.LogException(ex);
            }
            return false;
        }

        static void LoadConfig()
        {
            Logger.Instance.LogMessage("AutoOreAdmin.LoadConfig()");
            try
            {
                var worldname = MyAPIGateway.Session.Name;
                worldname = Regex.Replace(worldname, "[<>:\"/\\|?*]", "");  // Remove invalid filename chars

                System.IO.TextReader reader = MyAPIGateway.Utilities.ReadFileInLocalStorage(string.Format(ConfigFileName, worldname), typeof(AutoOreAdmin));
                var xmlData = reader.ReadToEnd();
                Configuration = MyAPIGateway.Utilities.SerializeFromXML<AutoOreConfig>(xmlData);
                reader.Close();

                Logger.Instance.Debug = Configuration.Debug;
            }
            catch( Exception ex)
            {
                Logger.Instance.LogMessage("AutoOreAdmin.LoadConfig() - Error");
                //Logger.Instance.LogException(ex);
                // ignore errors
            }
        }

        static public void SaveConfig(bool manual = false)
        {
            string filename = string.Empty;

            try
            {
                var worldname = MyAPIGateway.Session.Name;
                worldname = Regex.Replace(worldname, "[<>:\"/\\|?*]", "");  // Remove invalid filename chars

                var xmlData = MyAPIGateway.Utilities.SerializeToXML<AutoOreConfig>(Configuration);
                filename = MyAPIGateway.Utilities.GamePaths.ModsPath + "\\<modid>.sbm_AutoOreCollect\\" + string.Format(ConfigFileName, worldname);
                System.IO.TextWriter writer = MyAPIGateway.Utilities.WriteFileInLocalStorage(string.Format(ConfigFileName, worldname), typeof(AutoOreAdmin));
                writer.Write(xmlData);
                writer.Flush();
                writer.Close();

                if (manual)
                    MyAPIGateway.Utilities.ShowMessage(Globals.ModName, "Client config saved");
            }
            catch (Exception ex)
            {
                Logger.Instance.LogMessage(filename);
                Logger.Instance.LogException(ex);
            }

            // Send server save after client save, so client doesn't overwrite
            if (manual && !MyAPIGateway.Session.IsServer && MyAPIGateway.Session.Player.IsAdmin)
                MessageUtils.SendMessageToServer(new MessageSave());
        }
    }

    public static class AdminExtensions
    {
        /// <summary>
        /// Creates the objectbuilders in game, and syncs it to the server and all clients.
        /// </summary>
        /// <param name="entities"></param>
        public static void CreateAndSyncEntities(this List<VRage.ObjectBuilders.MyObjectBuilder_EntityBase> entities)
        {
            MyAPIGateway.Entities.RemapObjectBuilderCollection(entities);
            entities.ForEach(item => MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(item));
            MyAPIGateway.Multiplayer.SendEntitiesCreated(entities);
        }

    }

    //[Serializable]
    //public struct MyMyTuple
    //{
    //    public static MyTuple<T1, T2> Create<T1, T2>(T1 arg1, T2 arg2)
    //    {
    //        return new MyTuple<T1, T2>(arg1, arg2);
    //    }
    //}

    //[StructLayout(LayoutKind.Sequential, Pack = 4)]
    //[Serializable]
    //public struct MyTuple<T1, T2>
    //{
    //    public T1 Item1;
    //    public T2 Item2;

    //    public MyTuple(T1 item1, T2 item2)
    //    {
    //        Item1 = item1;
    //        Item2 = item2;
    //    }
    //}
}
// vim: tabstop=4 expandtab shiftwidth=4 nobackup
